Imports System.Threading  ' for threads
Imports System.Net.Sockets  ' for TcpClient and TcpServer

' This module contains two classes: ClientWinsock and ServerWinsock
Public Class ClientWinsock
    ' This object is created to connect to a TCPServer and to send a single Unicode message 
    Private ClientThread As Thread  ' used to run the main Sub StartClient of Client
    Private TCPIPPortNo As Integer ' TCPIP port number on destination computer
    Private DestinationComputer As String ' name or IP address of destination computer
    Private mMessage As String  ' Unicode message to be sent

    Public Sub New(ByVal Destination As String, ByVal Message As String)
        ' Split the given Message into Destination computer and TCPIP port number
        Dim pos As Integer
        ' Locate the character ";" in the Message string
        pos = Destination.IndexOf(";")
        If pos > 0 Then
            ' the part before ";" is the name or IP address of destination computer
            DestinationComputer = Destination.Substring(0, pos)
            TCPIPPortNo = CInt(Destination.Substring(pos + 1)) ' convert string to integer
        Else
            ' character ";" does not exist, that means only TCPIP Port number is given for Localhost
            DestinationComputer = "Localhost" ' Destination computer is Localhost
            TCPIPPortNo = CInt(Destination) ' convert string to integer
        End If
        mMessage = Message ' assign outgoing message to local string variable
        'Create a Thread object for Sub StartClient
        ClientThread = New Thread(AddressOf StartClient)
        'Starting the thread invokes the ThreadStart delegate
        ' i.e. run Sub StartClient in its own thread
        ClientThread.Start()
    End Sub

    Protected Sub StartClient()
        ' This is the main code in ClientWinsock. It's run in its own thread
        Dim Client As TcpClient ' Variable for the Client TCP socket
        Dim Buffer() As Byte ' used for outgoing message
        Try
            ' Instantiate TCPClient object
            Client = New TcpClient()
            ' Attempt to connect to destination (server) computer on given port number
            Client.Connect(DestinationComputer, TCPIPPortNo)
            ' Convert UFT8 message to an array of bytes before sending
            Buffer = System.Text.Encoding.UTF8.GetBytes(mMessage.ToCharArray)
            ' Send out the buffer
            Client.GetStream().Write(Buffer, 0, Buffer.Length)
            Client.Close()  ' Close the TcpClient
        Catch e As Exception
            ' Write to Console the message that cannot be sent
            Console.WriteLine("Can't send:" & mMessage)
        Finally
            ClientThread.Abort()  ' Abort thread
        End Try
    End Sub
End Class

Public Class ServerWinsock
    ' This object is created to serve many TCPClients and to receive Unicode messages 
    ' A thread is created to serve each TCPClient
    Const MaxThread As Integer = 500  ' Maximum number of threads that ServerWinsock can handle
    Private oListener As TcpListener ' Variable for TcpListener
    Private bStopListener As Boolean  ' Flag indicating that user wants to dispose this ServerWinsock
    Private ActiveThreads As Integer  ' Number of active threads, i.e. threads that are serving TCPClients
    ' Event that returns the incoming message
    Public Event OnMessage(ByVal IncomingMessage As String)

    Public Sub New(ByVal PortNo As Integer)
        ' Instantiate a TcpListener on given PortNo
        oListener = New TcpListener(PortNo)
        oListener.Start() ' Start the TcpListener
        ' Create a thread for Sub AcceptConnection
        Dim ServerThread As Thread
        ServerThread = New Thread(AddressOf AcceptConnection)
        ServerThread.Start() ' Run Sub AcceptConnection like a light weight child process
    End Sub

    Protected Sub ProcessRequest()
        Dim Buffer(5000) As Byte ' used to receive incoming message from TcpClient
        Dim bytes As Integer ' Actual number of bytes read
        Dim RecvMessage As String ' UTF8 text string of the buffer (array of bytes)
        ' Use oThread to reference the thread of this Sub
        Dim oThread As Thread
        oThread = Thread.CurrentThread()
        ' Accept request from the TcpClient
        Dim oSocket As Socket
        oSocket = oListener.AcceptSocket
        ' Keep looping until user wants to stop
        While Not bStopListener
            If oSocket.Available > 0 Then ' A message has arrived from TcpClient
                ' read the incoming message into a buffer
                bytes = oSocket.Receive(Buffer, Buffer.Length, 0)
                SyncLock oThread  ' Lock oThread
                    ' Convert the array of bytes (i.e. the buffer) to UTF8 text string
                    RecvMessage = System.Text.Encoding.UTF8.GetString(Buffer)
                    ' Raise an event to return the message to the program that owns this ServerWinsock
                    RaiseEvent OnMessage(RecvMessage)
                End SyncLock  ' unlock oThread
                Exit While
            End If
            ' get out of while loop if TcpClient has disconnected
            If Not oSocket.Connected Then Exit While
        End While
        oSocket.Close() ' Close the TcpServer socket
        SyncLock oThread ' Lock oThread
            ActiveThreads -= 1  ' Decrement number of Active Threads
        End SyncLock ' unlock oThread
    End Sub

    Private Sub AcceptConnection()
        'This is the main Sub of ServerWinsock
        ' Keep looping until user wants to stop
        Do While Not bStopListener
            Thread.Sleep(100)  ' Sleep 100 msec.
            If oListener.Pending() Then  ' received a request for connection from a TCPClient
                If ActiveThreads <= MaxThread Then
                    ' create a thread to handle this Client Request
                    Dim oThread As Thread
                    oThread = New Thread(AddressOf ProcessRequest)
                    oThread.Start()  ' Run Sub ProcessRequest
                    SyncLock oThread  ' Lock oThread so that ActiveThreads value is not changed while we're adding 1 to it
                        ActiveThreads += 1  ' Increment number of active threads
                    End SyncLock  ' Release the lock on oThread
                End If
            End If
        Loop
    End Sub

    Protected Overloads Sub Dispose(ByVal disposing As Boolean)
        bStopListener = True  ' Indicate user wants to dispose this ServerWinsock object
        oListener.Stop() ' Stop the TcpListener
    End Sub

End Class

